#include "scene.h"
#include "triangle.h"
#include "tmesh.h"
#include <float.h>

Scene *scene;

#include <iostream>

using namespace std;


Scene::Scene() {

  canvas = 0;
  canvasPPC = 0;
  refppc = 0;
  photoPPC = 0;
  photoR = 0;
  photoRPPC = 0;

  markersN = 4;
  markers = new V3[markersN];
  markers[0] = V3(82.0f, 174.0f, 0.0f);
  markers[1] = V3(281.0f, 152.0f, 0.0f);
  markers[2] = V3(628.0f, 265.0f, 0.0f);
  markers[3] = V3(293.0f, 374.0f, 0.0f);
    
  photo = 0;

  tms = 0;
  tmsN = 0;
  gui = new GUI();
  gui->show();

  int u0 = 20;
  int v0 = 50;
  int w = 648;
  int h = 484;
  fb = new FrameBuffer(u0, v0, w, h);
//  fb->label("SW framebuffer");
//  fb->show();
//  Setup();
//  ProjectiveTMSetup();
  PanoramaSetup();
  gui->uiw->position(canvas->w-200, canvas->h+100);
  Render();

}

void Scene::RenderExtrinsicCalibration() {

  fb->FrameSetup(0xFF00AAFF);
  if (photo)
    fb->Set(photo);

  unsigned int colors[4] = {0xFF0000FF, 0xFF00FF00, 0xFFFF0000, 0xFFFFFFFF};
  for (int i = 0; i < markersN; i++) {
    fb->DrawPoint((int)markers[i][0], (int)markers[i][1], 13, colors[i]);
  }

  for (int tmi = 0; tmi < tmsN; tmi++)
    tms[tmi]->RenderWireFrame(V3(1.0f, 0.0f, 0.0f), ppc, fb);

  fb->redraw();

}

void Scene::PanoramaRender() {

  canvas->Set(0xFF000000);

  canvas->Resample(canvasPPC, photo, photoPPC);
  canvas->Resample(canvasPPC, photoR, ppc);
  V3 p0 = ppc->Unproject(0, 0, 1.0f);
  V3 p1 = ppc->Unproject(0, ppc->h-1, 1.0f);
  V3 c0(1.0f, 0.0f, 0.0f), c1 = c0;
  canvas->DrawSegment3D(p0, p1, c0, c1, canvasPPC);

  ppc->Save("mydbg/currCamera.txt");

  canvas->redraw();

}

void Scene::Render() {

  if (canvas) {
    PanoramaRender();
    return;
  }

  fb->FrameSetup(0xFF00AAFF);

  for (int i = 0; i < tmsN; i++)
    tms[i]->RenderPTM(ppc, fb, photo, refppc);

//  tms[0]->RenderPTM(ppc, fb, photo, refppc);
//  tms[1]->RenderWireFrame(V3(1.0f, 0.0f, 0.0f), ppc, fb);

  fb->redraw();

}


void Scene::PanoramaSetup() {

  float hfov = 55.0f;
  photo = new FrameBuffer(0, 0, 10, 10);
  photo->Load("mydbg/pano1Small.tif");
  photoPPC = new PPC(photo->w, photo->h, hfov);

  photoR = new FrameBuffer(0, 0, 10, 10);
  photoR->Load("mydbg/pano0Small.tif");
  photoRPPC = new PPC(photoR->w, photoR->h, hfov);

  ppc = new PPC(photoR->w, photoR->h, hfov);

  float canvasHFOV = 90.0f;
  canvasPPC = new PPC(1200, 600, canvasHFOV);
  canvas = new FrameBuffer(0, 0, canvasPPC->w, canvasPPC->h);
  canvas->label("canvas");
  canvas->show();

}


void Scene::ProjectiveTMSetup() {

  float hfov = 55.0f;
  ppc = new PPC(fb->w, fb->h, hfov);
  photo = new FrameBuffer(0, 0, 10, 10);
  photo->Load("mydbg/photoScene4.tif");


  tms = new TMesh*[100];
  tms[tmsN] = new TMesh();
  tmsN++;
  V3 center(2.5f, 0.0f, -5.0f);
  V3 xaxis(1.0f, 0.0f, 0.0f);
  V3 yaxis(0.0f, 0.0f, 1.0f);
  float rw = 5.0f;
  float rh = 10.0f;
  tms[0]->SetRectangle(center, xaxis, yaxis, rw, rh);
  tms[0]->colors = new V3[4];
  tms[0]->colors[0] = V3(1.0f, 0.0f, 0.0f);
  tms[0]->colors[1] = V3(0.0f, 1.0f, 0.0f);
  tms[0]->colors[2] = V3(0.0f, 0.0f, 1.0f);
  tms[0]->colors[3] = V3(1.0f, 1.0f, 1.0f);

  tms[tmsN] = new TMesh();
  tmsN++;
  V3 c0(-80.0f/12.0f, -29.0f/12.0f, -10.0f - 45.0f/12.0f);
  V3 c1(5.0f+(90.0f+80.0f)/12.0f, (29.0f + 79.0f)/12.0f, 10.0f + (45.0f + 69.0f)/12.0f);
  c1 = c1 + c0;
  AABB aabb;
  aabb.AddPoint(c0);
  aabb.AddPoint(c1);
  tms[1]->SetAABB(aabb);

  refppc = new PPC(*ppc);
  refppc->Load("mydbg/bestCamera.txt");
  *ppc = *refppc;

  // render scene from photo camera view in order to get z-buffer
  Render();
  Fl::check();
  for (int i = 0; i < photo->w*photo->h; i++) {
    photo->zb[i] = fb->zb[i];
  }
#if 0
  unsigned int *oldpix = fb->pix;
  fb->pix = (unsigned int*)(photo->zb);
  fb->redraw();
  fb->Save("mydbg/dbg.tif");
  Fl::check();
  fb->pix = oldpix;
#endif

}

void Scene::Setup() {

  float hfov = 55.0f;
  ppc = new PPC(fb->w, fb->h, hfov);
  /*
  tm1 = new TMesh();
  tm1->Load("geometry/teapot1k.bin");
  tm1->SetCenter(V3(0.0f, 0.0f, -100.0f));
  tm1->Scale(100.0f);
  */

}

void Scene::DBG() {

  PanoRegister();
  return;


  ppc->Load("mydbg/bestCamera.txt");
  Render();
  return;

//  ppc->Save("mydbg/photoCamera.txt");
  ppc->Load("mydbg/photoCamera.txt");
  Render();
  Fl::check();
//  cerr << "INFO: initial guess error: " << ppc->ComputeError(tm1->verts, markers, 4) << " pix" << endl;
  ExtrinsicCalibrationBF();
  return;

  FrameBuffer *tmp = new FrameBuffer(0, 0, 10, 10);
  tmp->Load("mydbg/photoScene4.tif");
  tmp->label("loaded image");


//  tmp->DrawCircle(230.0f, 110.0f, 100.0f, 0xFF0000FF);

#if 0
  int u, v;
  for (v = 0; v < tmp->h/2; v++) {
    for (u = 0; u < tmp->w; u++) {
      unsigned int currColor = tmp->Get(u, v);
      unsigned int red = ((unsigned char*)&currColor)[0];
      red = 255-red;
      unsigned int green = ((unsigned char*)&currColor)[1];
      green = 255-green;
      unsigned int blue = ((unsigned char*)&currColor)[2];
      blue = 255-blue;
      currColor = 0xFF000000 + blue*256*256 + green*256 + red;
      tmp->Set(u, v, currColor);
    }
  }
#endif

  tmp->show();

  tmp->Save("mydbg/tmp.tif");


}


void Scene::ExtrinsicCalibrationBF() {

  // x, y, z, pan, tilt, roll
  int stepsN[6] = {15, 15, 15, 15, 15, 15};
  float amps[6] = {1.0f, 1.0f, 1.0f, 5.0f, 5.0f, 5.0f};
  int isteps[6];
  float minErr = FLT_MAX;
  float sol[6];
  PPC ppc0(*ppc);
  for (isteps[0] = 0; isteps[0] < stepsN[0]; isteps[0]++) {
    for (isteps[1] = 0; isteps[1] < stepsN[1]; isteps[1]++) {
      for (isteps[2] = 0; isteps[2] < stepsN[2]; isteps[2]++) {
        for (isteps[3] = 0; isteps[3] < stepsN[3]; isteps[3]++) {
          for (isteps[4] = 0; isteps[4] < stepsN[4]; isteps[4]++) {
            for (isteps[5] = 0; isteps[5] < stepsN[5]; isteps[5]++) {
              float currOffsets[6];
              for (int i = 0; i < 6; i++) {
                float frac = (float) isteps[i] / (float) (stepsN[i]-1);
                currOffsets[i] = -amps[i]/2.0f + amps[i]*frac;
              }
              PPC currPPC(*ppc);
              V3 currTrans(currOffsets[0], currOffsets[1], currOffsets[2]);
              currPPC.Translate(currTrans);
              currPPC.Pan(currOffsets[3]);
              currPPC.Tilt(currOffsets[4]);
              currPPC.Roll(currOffsets[5]);
              float currErr = currPPC.ComputeError(tms[0]->verts, markers, 4);
              if (currErr < minErr) {
                for (int i = 0; i < 6; i++) 
                  sol[i] = currOffsets[i];
                cerr << "INFO: found smaller error: " << currErr << "      \r";
                minErr = currErr;
                *ppc = currPPC;
                ppc->Save("mydbg/betterCamera.txt");
                Render();
                Fl::check();
                *ppc = ppc0;
              }
            }
          }
        }
      }
    }
  }
  cerr << endl << "INFO: DONE" << endl;

}

float Scene::PanoError(PPC *ppc0, FrameBuffer *fb0, PPC *ppc1, FrameBuffer *fb1) {

  float ret = 0.0f;
  int pixsN = 0;
  int du = 1;
  int dv = 1;
  for (int v = 0; v < ppc0->h; v+=dv) {
    for (int u = 0; u < ppc0->w; u+=du) {
      V3 P = ppc0->Unproject(u, v, 1.0f), projP;
      if (!ppc1->Project(P, projP))
        continue;
      unsigned int col1;
      if (fb1->Get(projP[0], projP[1], col1)) {
        V3 col1v; col1v.SetFromColor(col1);
        V3 col0v; col0v.SetFromColor(fb0->Get(u, v));
        ret += (col1v-col0v).Length();
        pixsN++;
      }
    }
  }

  int minPixsN = 1000;
  if (pixsN < minPixsN)
    return FLT_MAX;

  return ret / pixsN;

}

void Scene::PanoRegister() {

//  ppc->Load("mydbg/panoCameraGuess.txt");
  float err0 = PanoError(photoPPC, photo, ppc, photoR);
  cerr << "INFO: initial error: " << err0 << endl;
  Render();
  Fl::check();

  float minError = err0;
  V3 ptr(0.0f, 0.0f, 0.0f);
  PPC ppc0(*ppc);
  int panStepsN =27;
  int tiltStepsN = 13;
  int rollStepsN = 13;
  float panAmp = 41.0f;
  float tiltAmp = 7.0f;
  float rollAmp = 5.0f;
  int cstepi = 0;
  int cstepN = panStepsN * tiltStepsN * rollStepsN;
  for (int pi = 0; pi < panStepsN; pi++) {
    float pan = -1.0f * (float) pi / (float) (panStepsN-1) * panAmp;
    for (int ti = 0; ti < tiltStepsN; ti++) {
      float tilt = -tiltAmp / 2.0 + (float) ti / (float) (tiltStepsN-1) * tiltAmp;
      for (int ri = 0; ri < rollStepsN; ri++) {
        float roll = -rollAmp / 2.0 + (float) ri / (float) (rollStepsN-1) * rollAmp;
        *ppc = ppc0;
        ppc->Pan(pan);
        ppc->Tilt(tilt);
        ppc->Roll(roll);
        float currError = PanoError(photoPPC, photo, ppc, photoR);
        if (currError < minError) {
          minError = currError;
          ptr = V3(pan, tilt, roll);
          Render();
          Fl::check();
        }
        cstepi++;
        float pctg = (float)((int)(((float) cstepi / (float) cstepN)*10000.0f))/100.0f;
        if ((cstepi % 1) == 0) {
          cerr << "INFO: progress " << pctg << "%, error " << minError << "        \r";
        }
      }
    }
  }
  cerr << endl << "INFO: min error: " << minError << endl;
  cerr << "INFO: rot angles: " << ptr << endl;
  *ppc = ppc0;
  ppc->Pan(ptr[0]);
  ppc->Tilt(ptr[1]);
  ppc->Roll(ptr[2]);
  Render();
 
}

